module SM

  class Attr
    CHAR_BASE = ?a

    def initialize(number)
      @number = number
    end
    def as_char
      (CHAR_BASE + @number).chr
    end
    def as_bit
      1 << @number
    end
  end

  # Represent individual inline attributes. This class is
  # in two parts. The first is a set of class-level methods
  # that allow us to define attributes. The second is
  # the code that represents these attributes for a
  # particular case

  class AttrSet
    
    # Each attribute has
    # * a name (actually a symbol)
    # * a single character value, which is substituted into strings
    # * a bit mask, used when combining attributes in Attr objects

    class <<self
      def new_attribute(name)
        @attr_set    ||= {}
        @bit_to_name ||= {}
        unless @attr_set[name] 
          attr = Attr.new(@attr_set.size)
          bit  = attr.as_bit
          @attr_set[name]   = attr
          @bit_to_name[bit] = name
        end
      end
      def [](name)
        @attr_set[name].as_bit
      end
      def char(name)
        @attr_set[name].as_char
      end
      def name_from_bit(bit)
        @bit_to_name[bit]
      end
      def bit_from_char(char)
        1 << (char - Attr::CHAR_BASE)
      end
      def name_from_char(char)
        bit = bit_from_char(char)
        name_from_bit(bit_from_char(char))
      end
    end
    
    # -- Object stuff

    attr_reader :turn_on, :turn_off, :current


    def AttrSet.new_by_name(old_names, new_names)
      old_set = new_set = 0
      old_names.each {|n| old_set |= AttrSet[n]}
      new_names.each {|n| new_set |= AttrSet[n]}
      AttrSet.new(old_set, new_set)
    end

    
    def initialize(old_set, new_set)
      @turn_on  = new_set & ~old_set
      @turn_off = old_set & ~new_set
      @current  = new_set
    end

    def ==(other)
      @turn_on == other.turn_on &&
        @turn_off == other.turn_off &&
        @current  == other.current
    end

    def str_of(bits)
      obits = bits
      return "none" if bits == 0
      bit = 1
      res = []
      while bits != 0
        if (bits & 1) == 1
          res << AttrSet.name_from_bit(bit)
        end
        bits >>= 1
        bit <<= 1
      end
      obits.to_s + ":" + res.join(",")
    end
    
    def inspect
      "[AttrSet:+#{str_of(@turn_on)}/-#{str_of(@turn_off)}/=#{str_of(@current)}]"
    end
  end


  ##
  # Hold details of a special sequence

  class Special
    attr_reader :type, :text
    def initialize(type, text)
      @type, @text = type, text
    end
  end
  
  class AttributeManager

    ##
    # We work by substituting non-printing characters in to the
    # text. For now I'm assuming that I can substitute
    # a character in the range 0..8 for a 7 bit character
    # without damaging the encoded string, but this might
    # be optimistic
    #

    ATTR_FLAG  = 001
    A_START    = 002
    A_END      = 003
    A_PROTECT  = 004

    A_SPECIAL_START = 005
    A_SPECIAL_END   = 006

    START_ATTR   = ATTR_FLAG.chr + A_START.chr
    END_ATTR     = ATTR_FLAG.chr + A_END.chr

    START_SPECIAL = ATTR_FLAG.chr + A_SPECIAL_START.chr
    END_SPECIAL   = ATTR_FLAG.chr + A_SPECIAL_END.chr

    PROTECT_ATTR  = A_PROTECT.chr

    # This maps delimiters that occur around words (such as
    # *bold* or +tt+) where the start and end delimiters
    # and the same. This lets us optimize the regexp
    MATCHING_WORD_PAIRS = {}

    # And this is used when the delimiters aren't the same. In this
    # case the hash maps a pattern to the attribute character
    WORD_PAIR_MAP = {}

    # This maps HTML tags to the corresponding attribute char
    HTML_TAGS = {}

    # And tis maps _special_ sequences to a name. A specil sequence
    # in something like a WikiWord
    SPECIAL = {}

    # Return an attribute object with the given turn_on
    # and turn_off bits set

    def attribute(turn_on, turn_off)
      AttrSet.new_by_name(turn_on, turn_off)
    end

    # Return the bit corresponding to a particular attribute
    def bit_for(name)
      AttrSet[name]
    end

    # Map attributes like <b>text</b>to the sequence \001\002<char>\001\003<char>,
    # where <char> is a per-attribute specific character

    def convert_attrs
      # first do matching ones
      tags = MATCHING_WORD_PAIRS.keys.join("|")
      re = "([#{tags}])([A-Za-z]+?)\\1"
      @str.gsub!(Regexp.new(re)) {
        attr = MATCHING_WORD_PAIRS[$1];
        "#{START_ATTR}#{attr}#$2#{END_ATTR}#{attr}" 
      }

      # then non-matching
      unless WORD_PAIR_MAP.empty?
        WORD_PAIR_MAP.each do |regexp, attr|
          @str.gsub!(regexp) { "#{START_ATTR}#{attr}#$1#{END_ATTR}#{attr}" }
        end
      end
    end

    def convert_html
      tags = HTML_TAGS.keys.join("|")
      re = "<(#{tags})>(.*?)</\\1>"
      1 while @str.gsub!(Regexp.new(re, Regexp::IGNORECASE)) {
        attr = HTML_TAGS[$1.downcase];
        "#{START_ATTR}#{attr}#$2#{END_ATTR}#{attr}" 
      }
    end

    def convert_specials
      unless SPECIAL.empty?
        SPECIAL.each do |regexp, attr|
          @str.gsub!(regexp) { "#{START_SPECIAL}#{attr}#$1#{END_SPECIAL}#{attr}" }
        end
      end
    end

    # A \ in front of a character that would normally be
    # processed turns off processing. We do this by turning
    # \< into <#{PROTECT}
    
    PROTECTABLE = [ "<" << "\\" ]  #"


    def mask_protected_sequences
      protect_pattern = Regexp.new("\\\\([#{Regexp.escape(PROTECTABLE.join(''))}])")
      @str.gsub!(protect_pattern, "\\1#{PROTECT_ATTR}")
    end

    def unmask_protected_sequences
      @str.gsub!(/(.)#{PROTECT_ATTR}/, '\1')
    end

    def initialize
      add_word_pair("*", "*", :BOLD)
      add_word_pair("_", "_", :EM)
      add_word_pair("+", "+", :TT)
      
      add_html("em", :EM)
      add_html("i",  :EM)
      add_html("b",  :BOLD)
      add_html("tt", :TT)
    end

    def add_word_pair(start, stop, name)
      raise "Word flags may not start '<'" if start[0] == ?<
      AttrSet.new_attribute(name)
      if start == stop
        MATCHING_WORD_PAIRS[start] = AttrSet.char(name)
      else
        pattern = Regexp.new(Regexp.escape(start) + "([A-Za-z]+?)" + Regexp.escape(stop))
        WORD_PAIR_MAP[pattern] = AttrSet.char(name)
      end
      PROTECTABLE << start[0,1]
      PROTECTABLE.uniq!
    end

    def add_html(tag, name)
      AttrSet.new_attribute(name)
      HTML_TAGS[tag.downcase] = AttrSet.char(name)
    end

    def add_special(pattern, name)
      AttrSet.new_attribute(name)
      SPECIAL[pattern] = AttrSet.char(name)
    end

    def flow(str)
      @str = str
      puts("Before flow, str='#{@str.dump}'") if $DEBUG
      mask_protected_sequences
      convert_specials
      convert_attrs
      convert_html
      unmask_protected_sequences
      puts("After flow, str='#{@str.dump}'") if $DEBUG
      return split_into_flow
    end

    def split_into_flow
      current_attr = new_attr = 0
      # split into chunks of text and attribute sequences
      seq = @str.split(/(#{START_ATTR}.|#{END_ATTR}.|(?:#{START_SPECIAL}.(?:.*?)#{END_SPECIAL}.))/)

      # and interpret the attribute sequences
      result = []

      seq.map! do |entry|
        next if entry == ""
        if entry[0] == ATTR_FLAG
          bit = AttrSet.bit_from_char(entry[2])
          case entry[1]
          when A_START
            new_attr |= bit
          when A_END
            new_attr &= ~bit
          when A_SPECIAL_START
            type = AttrSet.name_from_char(entry[2])
            text = entry[3..-4]
	    if current_attr != new_attr
	      result << AttrSet.new(current_attr, new_attr)
	      current_attr = new_attr
	    end
            result << Special.new(type, text)
          else
            raise "Unknown attribute sequence: #{entry.dump}"
          end
        else
          if current_attr != new_attr
            result << AttrSet.new(current_attr, new_attr)
            current_attr = new_attr
          end
          result << entry
        end
      end

      result << AttrSet.new(current_attr, new_attr) if current_attr != new_attr
      result
    end

  end

end
