class Container
  def initialize
    @entries = []
  end
  def <<(entry)
    @entries << entry
  end
  def each(&block)
    @entries.each(&block)
  end
  def to_s
    @entries.collect{|e| e.to_s }.join("\n====\n")
  end
end

module Viewable
  def viewer
    @viewer
  end
  def viewer=(v)
    @viewer = v
  end
end

class Context
  attr_reader :name, :classes, :modules, :method_list, :attributes, :aliases
  attr_reader :requires, :includes
  attr_accessor :parent
  attr_accessor   :comment

  include Viewable

  def initialize
    @classes = []
    @modules = []
    @method_list = []
    @attributes = []
    @aliases = []
    @requires = []
    @includes = []

    @name    ||= "unknown"
    @comment ||= ""

    @visibility = :public
  end

  def ongoing_visibility=(vis)
    @visibility = vis
  end

  # Given an array +methods+ of method names, set the
  # visibility of the corresponding AnyMethod object

  def set_visibility_for(methods, vis)
    @method_list.each_with_index do |m,i|
      if methods.include?(m.name)
        m.visibility = vis
      end
    end
  end


  def add_class(a_class)
    puts "Adding class #{a_class.name} to #@name" if $DEBUG
    add_to(@classes, a_class)
  end

  def add_module(a_module)
    puts "Adding module #{a_module.name} to #@name" if $DEBUG
    add_to(@modules, a_module)
  end

  def add_method(a_method)
    puts "Adding method #{a_method.name} to #@name" if $DEBUG
    a_method.visibility = @visibility
    add_to(@method_list, a_method)
  end

  def add_attribute(an_attribute)
    add_to(@attributes, an_attribute)
  end

  def add_alias(an_alias)
    add_to(@aliases, an_alias)
  end

  def add_include(an_include)
    add_to(@includes, an_include)
  end
  
  # Requires always get added to the top-level (file) context
  def add_require(a_require)
    if self.kind_of? TopLevel
      add_to(@requires, a_require)
    else
      parent.add_require(a_require)
    end
  end

  def add_to(array, thing)
    array << thing
    thing.parent = self
  end

  # Iterate over all the clesses and modules in
  # this object

  def each_classmodule
    @modules.each {|m| yield m}
    @classes.each {|c| yield c}
  end

  def each_method
    @method_list.each {|m| yield m}
  end

  def each_attribute 
    @attributes.each {|a| yield a}
  end

  # Return the toplevel that owns us

  def toplevel
    return @toplevel if defined? @toplevel
    @toplevel = self
    @toplevel = @toplevel.parent until TopLevel === @toplevel
    @toplevel
  end
end

class TopLevel < Context
  attr_accessor :file_stat
  attr_accessor :file_relative_name
  attr_accessor :file_absolute_name

  def initialize
    super
    @name = "TopLevel"
  end

  def full_name
    nil
  end
end

class ClassModule < Context

  attr_accessor :superclass

  def initialize(text, name, tk, comment, superclass = nil)
    @text = text
    @name = name
    @superclass = superclass
    @token = tk
    @comment = comment
    super()
  end

  def full_name
    if @parent && @parent.full_name
      @parent.full_name + "::" + @name
    else
      @name
    end
  end

  def to_s
    res = self.class.name + ": " + @name + " (" + @text + ")\n"
    res << @comment.to_s
    
    res << super
    
    res
  end

  def is_module?
    false
  end
end

class AnonClass < ClassModule
end

class NormalClass < ClassModule
end

class SingleClass < ClassModule
end

class NormalModule < ClassModule
  def is_module?
    true
  end
end


class AnyMethod
  attr_reader :name, :comment
  attr_accessor :parent
  attr_accessor :params
  attr_accessor :visibility
  attr_accessor :block_params

  include Viewable

  def initialize(text, name, tk, comment)
    @text = text
    @name = name
    @comment = comment
    @token = tk
    @visibility = :public
  end

  def <=>(other)
    @name <=> other.name
  end

  def to_s
    res = self.class.name + ": " + @name + " (" + @text + ")\n"
    res << @comment.to_s
    res
  end
end

class SingleMethod < AnyMethod
  def initialize(*args)
    super
  end

  def singleton?
    true
  end
end

class NormalMethod < AnyMethod
  def initialize(*args)
    super
  end

  def singleton?
    false
  end
end

Alias = Struct.new(:text, :old_name, :new_name, :tk, :comment)
class Alias
  attr_accessor :parent

  include Viewable

  def to_s
    "alias: #{self.old_name} ->  #{self.new_name}\n#{self.comment}"
  end
end

Attr = Struct.new(:text, :name, :rw, :tk, :comment)
class Attr
  attr_accessor :parent

  include Viewable

  def to_s
    "attr: #{self.name} #{self.rw}\n#{self.comment}"
  end

  def <=>(other)
    self.name <=> other.name
  end
end

Require = Struct.new(:name, :comment)
class Require
  attr_accessor :parent
end

Include = Struct.new(:name, :comment)
class Include
  attr_accessor :parent
end
